open Absyn
open Format 

exception TODO
exception RuntimeError of string

type value = 
	 Num of int 
  | ProcVal of env * var list * exp 
  | Bool of bool
and  env = Extend of var list * value array * env 
			| ExtendRec of var list * var list list * exp list * env
			| Empty

let rec print_value (v:value) : unit = 
  match v with 
		Num i -> Format.print_int i;
	 | Bool b -> if b then Format.print_string "true" else 
		  Format.print_string "false"
	 | ProcVal (env,v,str) -> Format.print_string "<closure>"

let closure (ids:var list) (body:exp) (env:env) : value = 
  ProcVal (env, ids, body)

(* ------- Environments --------------- *)

let list_find_position (x:'a) (l:'a list) :int option = 
  let rec helper l i = 
	 match l with 
		  [] -> None
		| (hd::tl) -> if x = hd then Some i else 
			 helper tl (i+1)
  in
	 helper l 0

let extend_env (vars:var list) (vals:value list) (old_env:env) : env = 
  Extend(vars, Array.of_list vals, old_env)

let extend_env_rec (proc_names:var list) (idss:var list list) (bodies:exp list) (old_env:env) : env = 
  ExtendRec (proc_names, idss, bodies, old_env)

				
let rec apply_env (env:env) (sym:var) :value = 
  match env with 
		Empty -> raise (RuntimeError ("Unbound variable " ^ sym))
	 | Extend (vars, vals, old_env) -> 
		  (match list_find_position sym vars with 
				 Some pos -> (try 
									 vals.(pos)
								  with _ -> 
									 raise (RuntimeError ("Unbound variable " ^ sym)))
			  | None -> apply_env old_env sym)
	 | ExtendRec (proc_names, idss, bodies, old_env) ->
		  match (list_find_position sym proc_names) with 
				Some pos -> 
				  closure (List.nth idss pos) (List.nth bodies pos) env
			 | None -> apply_env old_env sym


(* ----- Booleans ------ *)

let is_true (v:value) :bool = 
  match v with 
		Num i -> not (i=0)
	 | _ -> false
		 
let true_value  :value = Num 1
let false_value :value = Num 0
 
(* ----------- Primitives ----------- *)

let rec apply_prim (p:prim) (args:value list) = 
  match p,args with 
		Plus,  [Num v1;Num v2] -> Num (v1 + v2)
	 | Times, [Num v1;Num v2] -> Num (v1 * v2)
	 | Minus, [Num v1;Num v2] -> Num (v1 - v2)
	 | And,   [v1;v2] ->
		  if (is_true v1 && is_true v2)
		  then true_value
		  else false_value
	 | Or,    [v1;v2] ->
		  if (is_true v1 || is_true v2)
		  then true_value
		  else false_value
	 | Not,   [v1] ->
		  if (is_true v1) 
		  then false_value 
		  else true_value
	 | Add1, [Num v1] -> Num (v1 + 1)
	 | Sub1, [Num v1] -> Num (v1 - 1)
	 | _, _ -> raise (RuntimeError "Unimplemented primitive or invalid arguments")

(* ------------ Evaluator ---------------------- *)

let rec apply_procval (v:value) (args:value list) :value = 
  match v with 
		ProcVal (env, ids, body) -> 
		  eval body (extend_env ids args env)
	 | _ -> raise (RuntimeError "Attempt to apply non-procedure")

and  eval (exp:exp) (env:env) :value = 
  match exp with 
		LitInt i -> Num i
	 | LitBool b -> 
		  if b then true_value 
		  else false_value
	 | Var v -> apply_env env v
	 | Prim (p, rands) -> 
		  let args = eval_rands rands env in
			 apply_prim p args
	 | If (test_exp, true_exp, false_exp) -> 
		  if (is_true (eval test_exp env))
		  then (eval true_exp env)
		  else (eval false_exp env)
	 | Let (vars,rands,body) -> 
		  let args = eval_rands rands env in 
			 eval body (extend_env vars args env)
	 | Lambda (ids,body) -> 		
		  closure ids body env
	 | App (rator,rands) -> 
		  let proc = eval rator env in 
		  let args = eval_rands rands env in 
			 apply_procval proc args 
	 | Letrec (proc_names, idss, bodies, letrec_body) -> 
		  eval letrec_body (extend_env_rec proc_names idss bodies env)
				  
and eval_rands (rands:exp list) (env:env): value list = 
  List.map (fun x -> eval x env) rands

		

